# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Propose command implementations."""

from io import StringIO

from ... import branch as _mod_branch
from ... import controldir, errors, msgeditor, urlutils
from ... import forge as _mod_forge
from ... import log as _mod_log
from ... import missing as _mod_missing
from ...commands import Command
from ...i18n import gettext
from ...option import ListOption, Option, RegistryOption
from ...trace import note, warning


def branch_name(branch):
    if branch.name:
        return branch.name
    return urlutils.basename(branch.user_url)


def _check_already_merged(branch, target):
    # TODO(jelmer): Check entire ancestry rather than just last revision?
    if branch.last_revision() == target.last_revision():
        raise errors.CommandError(
            gettext("All local changes are already present in target.")
        )


class cmd_publish_derived(Command):
    __doc__ = """Publish a derived branch.

    Try to create a public copy of a local branch on a hosting site,
    derived from the specified base branch.

    Reasonable defaults are picked for owner name, branch name and project
    name, but they can also be overridden from the command-line.
    """

    takes_options = [
        "directory",
        Option("owner", help="Owner of the new remote branch.", type=str),
        Option("project", help="Project name for the new remote branch.", type=str),
        Option("name", help="Name of the new remote branch.", type=str),
        Option("no-allow-lossy", help="Allow fallback to lossy push, if necessary."),
        Option("overwrite", help="Overwrite existing commits."),
        "revision",
    ]
    takes_args = ["submit_branch?"]

    def run(
        self,
        submit_branch=None,
        owner=None,
        name=None,
        project=None,
        no_allow_lossy=False,
        overwrite=False,
        directory=".",
        revision=None,
    ):
        local_branch = _mod_branch.Branch.open_containing(directory)[0]
        self.add_cleanup(local_branch.lock_write().unlock)
        if submit_branch is None:
            submit_branch = local_branch.get_submit_branch()
            note(gettext("Using submit branch %s") % submit_branch)
        if submit_branch is None:
            submit_branch = local_branch.get_parent()
            note(gettext("Using parent branch %s") % submit_branch)
        submit_branch = _mod_branch.Branch.open(submit_branch)
        _check_already_merged(local_branch, submit_branch)
        if name is None:
            name = branch_name(local_branch)
        forge = _mod_forge.get_forge(submit_branch)
        if revision is None:
            stop_revision = None
        else:
            stop_revision = revision.as_revision_id(branch)
        remote_branch, public_url = forge.publish_derived(
            local_branch,
            submit_branch,
            name=name,
            project=project,
            owner=owner,
            allow_lossy=not no_allow_lossy,
            overwrite=overwrite,
            revision_id=stop_revision,
        )
        local_branch.set_push_location(remote_branch.user_url)
        local_branch.set_public_branch(public_url)
        local_branch.set_submit_branch(submit_branch.user_url)
        note(gettext("Pushed to %s") % public_url)


def summarize_unmerged(local_branch, remote_branch, target, prerequisite_branch=None):
    """Generate a text description of the unmerged revisions in branch.

    :param branch: The proposed branch
    :param target: Target branch
    :param prerequisite_branch: Optional prerequisite branch
    :return: A string
    """
    log_format = _mod_log.log_formatter_registry.get_default(local_branch)
    to_file = StringIO()
    lf = log_format(to_file=to_file, show_ids=False, show_timezone="original")
    if prerequisite_branch:
        local_extra = _mod_missing.find_unmerged(
            remote_branch, prerequisite_branch, restrict="local"
        )[0]
    else:
        local_extra = _mod_missing.find_unmerged(
            remote_branch, target, restrict="local"
        )[0]

    if remote_branch.supports_tags():
        rev_tag_dict = remote_branch.tags.get_reverse_tag_dict()
    else:
        rev_tag_dict = {}

    for revision in _mod_missing.iter_log_revisions(
        local_extra, local_branch.repository, False, rev_tag_dict
    ):
        lf.log_revision(revision)
    return to_file.getvalue()


class cmd_propose_merge(Command):
    __doc__ = """Propose a branch for merging.

    This command creates a merge proposal for the local
    branch to the target branch. The format of the merge
    proposal depends on the submit branch.
    """

    takes_options = [
        "directory",
        RegistryOption(
            "forge", help="Use the forge.", lazy_registry=("breezy.forge", "forges")
        ),
        ListOption("reviewers", short_name="R", type=str, help="Requested reviewers."),
        Option("name", help="Name of the new remote branch.", type=str),
        Option("description", help="Description of the change.", type=str),
        Option("prerequisite", help="Prerequisite branch.", type=str),
        Option("wip", help="Mark merge request as work-in-progress"),
        Option("auto", help="Automatically merge when the CI passes"),
        Option(
            "commit-message",
            help="Set commit message for merge, if supported",
            type=str,
        ),
        ListOption("labels", short_name="l", type=str, help="Labels to apply."),
        Option("no-allow-lossy", help="Allow fallback to lossy push, if necessary."),
        Option(
            "allow-collaboration",
            help="Allow collaboration from target branch maintainer(s)",
        ),
        Option("allow-empty", help="Do not prevent empty merge proposals."),
        Option("overwrite", help="Overwrite existing commits."),
        Option("open", help="Open merge proposal in web browser"),
        Option(
            "delete-source-after-merge",
            help="Delete source branch when proposal is merged",
        ),
        "revision",
    ]
    takes_args = ["submit_branch?"]

    aliases = ["propose"]

    def run(
        self,
        submit_branch=None,
        directory=".",
        forge=None,
        reviewers=None,
        name=None,
        no_allow_lossy=False,
        description=None,
        labels=None,
        prerequisite=None,
        commit_message=None,
        wip=False,
        allow_collaboration=False,
        allow_empty=False,
        overwrite=False,
        open=False,
        auto=False,
        delete_source_after_merge=None,
        revision=None,
    ):
        tree, branch, relpath = controldir.ControlDir.open_containing_tree_or_branch(
            directory
        )
        if submit_branch is None:
            submit_branch = branch.get_submit_branch()
        if submit_branch is None:
            submit_branch = branch.get_parent()
        if submit_branch is None:
            raise errors.CommandError(
                gettext("No target location specified or remembered")
            )
        target = _mod_branch.Branch.open(submit_branch)
        if not allow_empty:
            _check_already_merged(branch, target)
        if forge is None:
            forge = _mod_forge.get_forge(target)
        else:
            forge = forge.probe(target)
        if name is None:
            name = branch_name(branch)
        if revision is None:
            stop_revision = None
        else:
            stop_revision = revision.as_revision_id(branch)
        remote_branch, public_branch_url = forge.publish_derived(
            branch,
            target,
            name=name,
            allow_lossy=not no_allow_lossy,
            overwrite=overwrite,
            revision_id=stop_revision,
        )
        branch.set_push_location(remote_branch.user_url)
        branch.set_submit_branch(target.user_url)
        note(
            gettext("Published branch to %s"),
            forge.get_web_url(remote_branch) or public_branch_url,
        )
        if prerequisite is not None:
            prerequisite_branch = _mod_branch.Branch.open(prerequisite)
        else:
            prerequisite_branch = None
        proposal_builder = forge.get_proposer(remote_branch, target)
        if description is None:
            body = proposal_builder.get_initial_body()
            info = proposal_builder.get_infotext()
            info += "\n\n" + summarize_unmerged(
                branch, remote_branch, target, prerequisite_branch
            )
            description = msgeditor.edit_commit_message(info, start_message=body)
        try:
            proposal = proposal_builder.create_proposal(
                description=description,
                reviewers=reviewers,
                prerequisite_branch=prerequisite_branch,
                labels=labels,
                commit_message=commit_message,
                work_in_progress=wip,
                allow_collaboration=allow_collaboration,
                delete_source_after_merge=delete_source_after_merge,
            )
        except _mod_forge.MergeProposalExists as e:
            note(gettext("There is already a branch merge proposal: %s"), e.url)
        else:
            note(gettext("Merge proposal created: %s") % proposal.url)
            if open:
                web_url = proposal.get_web_url()
                import webbrowser

                note(gettext("Opening %s in web browser"), web_url)
                webbrowser.open(web_url)
            if auto:
                proposal.merge(auto=True)


class cmd_find_merge_proposal(Command):
    __doc__ = """Find a merge proposal.

    """

    takes_options = ["directory"]
    takes_args = ["submit_branch?"]
    aliases = ["find-proposal"]

    def run(self, directory=".", submit_branch=None):
        tree, branch, relpath = controldir.ControlDir.open_containing_tree_or_branch(
            directory
        )
        public_location = branch.get_public_branch()
        if public_location:
            branch = _mod_branch.Branch.open(public_location)
        if submit_branch is None:
            submit_branch = branch.get_submit_branch()
        if submit_branch is None:
            submit_branch = branch.get_parent()
        if submit_branch is None:
            raise errors.CommandError(
                gettext("No target location specified or remembered")
            )
        else:
            target = _mod_branch.Branch.open(submit_branch)
        forge = _mod_forge.get_forge(branch)
        for mp in forge.iter_proposals(branch, target):
            self.outf.write(gettext("Merge proposal: %s\n") % mp.url)


class cmd_my_merge_proposals(Command):
    __doc__ = """List all merge proposals owned by the logged-in user.

    """

    hidden = True

    takes_args = ["base_url?"]
    takes_options = [
        "verbose",
        RegistryOption.from_kwargs(
            "status",
            title="Proposal Status",
            help="Only include proposals with specified status.",
            value_switches=True,
            enum_switch=True,
            all="All merge proposals",
            open="Open merge proposals",
            merged="Merged merge proposals",
            closed="Closed merge proposals",
        ),
        RegistryOption(
            "forge", help="Use the forge.", lazy_registry=("breezy.forge", "forges")
        ),
    ]

    def run(self, status="open", verbose=False, forge=None, base_url=None):
        for instance in _mod_forge.iter_forge_instances(forge=forge):
            if base_url is not None and instance.base_url != base_url:
                continue
            try:
                for mp in instance.iter_my_proposals(status=status):
                    self.outf.write("{}\n".format(mp.url))
                    if verbose:
                        source_branch_url = mp.get_source_branch_url()
                        if source_branch_url:
                            self.outf.write(
                                "(Merging {} into {})\n".format(
                                    source_branch_url, mp.get_target_branch_url()
                                )
                            )
                        else:
                            self.outf.write(
                                "(Merging into {})\n".format(mp.get_target_branch_url())
                            )
                        description = mp.get_description()
                        if description:
                            self.outf.writelines(
                                ["\t{}\n".format(l) for l in description.splitlines()]
                            )
                        self.outf.write("\n")
            except _mod_forge.ForgeLoginRequired:
                warning("Skipping %s, login required.", instance)


class cmd_land_merge_proposal(Command):
    __doc__ = """Land a merge proposal."""

    takes_args = ["url"]
    takes_options = [Option("message", help="Commit message to use.", type=str)]

    def run(self, url, message=None):
        proposal = _mod_forge.get_proposal_by_url(url)
        proposal.merge(commit_message=message)


class cmd_web_open(Command):
    __doc__ = """Open a branch page in your web browser."""

    takes_options = [
        Option(
            "dry-run",
            "Do not actually open the browser. Just say the URL we would use.",
        ),
    ]
    takes_args = ["location?"]

    def _possible_locations(self, location):
        """Yield possible external locations for the branch at 'location'."""
        yield location
        try:
            branch = _mod_branch.Branch.open_containing(location)[0]
        except errors.NotBranchError:
            return
        branch_url = branch.get_public_branch()
        if branch_url is not None:
            yield branch_url
        branch_url = branch.get_push_location()
        if branch_url is not None:
            yield branch_url

    def _get_web_url(self, location):
        for branch_url in self._possible_locations(location):
            try:
                branch = _mod_branch.Branch.open_containing(branch_url)[0]
            except errors.NotBranchError as e:
                mutter("Unable to open branch %s: %s", branch_url, e)
                continue

            try:
                forge = _mod_forge.get_forge(branch)
            except _mod_forge.UnsupportedForge:
                continue

            return forge.get_web_url(branch)
        raise errors.CommandError("Unable to get web URL for {}".format(location))

    def run(self, location=None, dry_run=False):
        if location is None:
            location = "."
        web_url = self._get_web_url(location)
        note(gettext("Opening %s in web browser") % web_url)
        if not dry_run:
            import webbrowser

            # otherwise brz.exe lacks this module
            webbrowser.open(web_url)


class cmd_forges(Command):
    __doc__ = """List all known hosting sites and user details."""

    hidden = True

    def run(self):
        for instance in _mod_forge.iter_forge_instances():
            current_user = instance.get_current_user()
            if current_user is not None:
                current_user_url = instance.get_user_url(current_user)
                if current_user_url is not None:
                    self.outf.write(
                        gettext("%s (%s) - user: %s (%s)\n")
                        % (
                            instance.name,
                            instance.base_url,
                            current_user,
                            current_user_url,
                        )
                    )
                else:
                    self.outf.write(
                        gettext("%s (%s) - user: %s\n")
                        % (instance.name, instance.base_url, current_user)
                    )
            else:
                self.outf.write(
                    gettext("%s (%s) - not logged in\n")
                    % (instance.name, instance.base_url)
                )



#!/cygdrive/C/Python25/python
"""A script to help automate the build process."""

# When preparing a new release, make sure to set all of these to the latest
# values.
VERSIONS = {
    "brz": "1.17",
    "qbzr": "0.12",
    "bzrtools": "1.17.0",
    "bzr-svn": "0.6.3",
    "bzr-rewrite": "0.5.2",
    "subvertpy": "0.6.8",
}

# This will be passed to 'make' to ensure we build with the right python
PYTHON = "/cygdrive/c/Python25/python"

# Create the final build in this directory
TARGET_ROOT = "release"

DEBUG_SUBPROCESS = True


import os
import shutil
import subprocess
import sys

BRZ_EXE = None


def brz():
    global BRZ_EXE
    if BRZ_EXE is not None:
        return BRZ_EXE
    try:
        subprocess.call(
            ["brz", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        BRZ_EXE = "brz"
    except OSError:
        try:
            subprocess.call(
                ["brz.bat", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            BRZ_EXE = "brz.bat"
        except OSError:
            raise RuntimeError("Could not find brz or brz.bat on your path.")
    return BRZ_EXE


def call_or_fail(*args, **kwargs):
    """Call a subprocess, and fail if the return code is not 0."""
    if DEBUG_SUBPROCESS:
        print('  calling: "{}"'.format(" ".join(args[0])))
    p = subprocess.Popen(*args, **kwargs)
    (out, err) = p.communicate()
    if p.returncode != 0:
        raise RuntimeError("Failed to run: {}, {}".format(args, kwargs))
    return out


TARGET = None


def get_target():
    global TARGET
    if TARGET is not None:
        return TARGET
    out = call_or_fail(
        [get_brz_dir() + "/brz", "version", "--short"], stdout=subprocess.PIPE
    )
    version = out.strip()
    TARGET = os.path.abspath(TARGET_ROOT + "-" + version)
    return TARGET


def clean_target():
    """Nuke the target directory so we know we are starting from scratch."""
    target = get_target()
    if os.path.isdir(target):
        print("Deleting: {}".format(target))
        shutil.rmtree(target)


def get_brz_dir():
    return "brz." + VERSIONS["brz"]


def update_brz():
    """Make sure we have the latest brz in play."""
    brz_dir = get_brz_dir()
    if not os.path.isdir(brz_dir):
        brz_version = VERSIONS["brz"]
        brz_url = "lp:brz/" + brz_version
        print("Getting brz release {} from {}".format(brz_version, brz_url))
        call_or_fail([brz(), "co", brz_url, brz_dir])
    else:
        print("Ensuring {} is up-to-date".format(brz_dir))
        call_or_fail([brz(), "update", brz_dir])


def create_target():
    target = get_target()
    print("Creating target dir: {}".format(target))
    call_or_fail([brz(), "co", get_brz_dir(), target])


def get_plugin_trunk_dir(plugin_name):
    return "{}/trunk".format(plugin_name)


def get_plugin_release_dir(plugin_name):
    return "{}/{}".format(plugin_name, VERSIONS[plugin_name])


def get_plugin_trunk_branch(plugin_name):
    return "lp:{}".format(plugin_name)


def update_plugin_trunk(plugin_name):
    trunk_dir = get_plugin_trunk_dir(plugin_name)
    if not os.path.isdir(trunk_dir):
        plugin_trunk = get_plugin_trunk_branch(plugin_name)
        print("Getting latest {} trunk".format(plugin_name))
        call_or_fail([brz(), "co", plugin_trunk, trunk_dir])
    else:
        print("Ensuring {} is up-to-date".format(trunk_dir))
        call_or_fail([brz(), "update", trunk_dir])
    return trunk_dir


def _plugin_tag_name(plugin_name):
    if plugin_name in ("bzr-svn", "bzr-rewrite", "subvertpy"):
        return "{}-{}".format(plugin_name, VERSIONS[plugin_name])
    # bzrtools and qbzr use 'release-X.Y.Z'
    return "release-" + VERSIONS[plugin_name]


def update_plugin(plugin_name):
    release_dir = get_plugin_release_dir(plugin_name)
    if not os.path.isdir(plugin_name):
        if plugin_name in ("bzr-svn", "bzr-rewrite"):
            # bzr-svn uses a different repo format
            call_or_fail([brz(), "init-shared-repo", "--rich-root-pack", plugin_name])
        else:
            os.mkdir(plugin_name)
    if os.path.isdir(release_dir):
        print("Removing existing dir: {}".format(release_dir))
        shutil.rmtree(release_dir)
    # First update trunk
    trunk_dir = update_plugin_trunk(plugin_name)
    # Now create the tagged directory
    tag_name = _plugin_tag_name(plugin_name)
    print("Creating the branch {}".format(release_dir))
    call_or_fail([brz(), "co", "-rtag:{}".format(tag_name), trunk_dir, release_dir])
    return release_dir


def install_plugin(plugin_name):
    release_dir = update_plugin(plugin_name)
    # at least bzrtools doesn't like you to call 'setup.py' unless you are in
    # that directory specifically, so we cd, rather than calling it from
    # outside
    print("Installing {}".format(release_dir))
    call_or_fail(
        [
            sys.executable,
            "setup.py",
            "install",
            "-O1",
            "--install-lib={}".format(get_target()),
        ],
        cwd=release_dir,
    )


def update_tbzr():
    tbzr_loc = os.environ.get("TBZR", None)
    if tbzr_loc is None:
        raise ValueError("You must set TBZR to the location of tortoisebzr.")
    print("Updating {}".format(tbzr_loc))
    call_or_fail([brz(), "update", tbzr_loc])


def build_installer():
    target = get_target()
    print()
    print()
    print("*" * 60)
    print("Building standalone installer")
    call_or_fail(["make", "PYTHON={}".format(PYTHON), "installer"], cwd=target)


def main(args):
    import optparse

    p = optparse.OptionParser(usage="%prog [OPTIONS]")
    opts, args = p.parse_args(args)

    update_brz()
    update_tbzr()
    clean_target()
    create_target()
    install_plugin("subvertpy")
    install_plugin("bzrtools")
    install_plugin("qbzr")
    install_plugin("bzr-svn")
    install_plugin("bzr-rewrite")

    build_installer()


if __name__ == "__main__":
    main(sys.argv[1:])

# vim: ts=4 sw=4 sts=4 et ai



# Copyright (C) 2006-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import errors, tests, workingtree


class TestWorkingTree(tests.TestCaseWithTransport):
    def make_branch_and_tree(self, relpath, format=None):
        if format is None:
            format = "development-subtree"
        return tests.TestCaseWithTransport.make_branch_and_tree(self, relpath, format)

    def make_trees(self, format=None, same_root=False):
        self.build_tree(["tree/", "tree/file", "tree/subtree/", "tree/subtree/file2"])
        base_tree = self.make_branch_and_tree("tree", format=format)
        base_tree.add("file", ids=b"file-id")
        base_tree.commit("first commit", rev_id=b"tree-1")
        sub_tree = self.make_branch_and_tree(
            "tree/subtree", format="development-subtree"
        )
        if same_root is True:
            sub_tree.set_root_id(base_tree.path2id(""))
        sub_tree.add("file2", ids=b"file2-id")
        sub_tree.commit("first commit", rev_id=b"subtree-1")
        return base_tree, sub_tree

    def test_old_knit1_failure(self):
        """Ensure that BadSubsumeSource is raised.

        SubsumeTargetNeedsUpgrade must not be raised, because upgrading the
        target won't help.
        """
        base_tree, sub_tree = self.make_trees(format="knit", same_root=True)
        self.assertRaises(errors.BadSubsumeSource, base_tree.subsume, sub_tree)

    def test_knit1_failure(self):
        base_tree, sub_tree = self.make_trees(format="knit")
        self.assertRaises(errors.SubsumeTargetNeedsUpgrade, base_tree.subsume, sub_tree)

    def test_subsume_tree(self):
        base_tree, sub_tree = self.make_trees()
        self.assertNotEqual(base_tree.path2id(""), sub_tree.path2id(""))
        sub_root_id = sub_tree.path2id("")
        # this test checks the subdir is removed, so it needs to know the
        # control directory; that changes rarely so just hardcode (and check)
        # it is correct.
        self.assertPathExists("tree/subtree/.bzr")
        base_tree.subsume(sub_tree)
        self.assertEqual([b"tree-1", b"subtree-1"], base_tree.get_parent_ids())
        self.assertEqual(sub_root_id, base_tree.path2id("subtree"))
        self.assertEqual(b"file2-id", base_tree.path2id("subtree/file2"))
        # subsuming the tree removes the control directory, so you can't open
        # it.
        self.assertPathDoesNotExist("tree/subtree/.bzr")
        with open("tree/subtree/file2", "rb") as file2:
            file2_contents = file2.read()
        base_tree = workingtree.WorkingTree.open("tree")
        base_tree.commit("combined", rev_id=b"combined-1")
        self.assertEqual(b"file2-id", base_tree.path2id("subtree/file2"))
        if base_tree.supports_setting_file_ids():
            self.assertEqual("subtree/file2", base_tree.id2path(b"file2-id"))
        self.assertEqualDiff(file2_contents, base_tree.get_file_text("subtree/file2"))
        basis_tree = base_tree.basis_tree()
        basis_tree.lock_read()
        self.addCleanup(basis_tree.unlock)
        self.assertEqualDiff(file2_contents, base_tree.get_file_text("subtree/file2"))
        self.assertEqualDiff(file2_contents, basis_tree.get_file_text("subtree/file2"))
        self.assertEqual(b"subtree-1", basis_tree.get_file_revision("subtree/file2"))
        self.assertEqual(b"combined-1", basis_tree.get_file_revision("subtree"))

    def test_subsume_failure(self):
        base_tree, sub_tree = self.make_trees()
        if base_tree.path2id("") == sub_tree.path2id(""):
            raise tests.TestSkipped("This test requires unique roots")
        self.assertRaises(errors.BadSubsumeSource, base_tree.subsume, base_tree)
        self.assertRaises(errors.BadSubsumeSource, sub_tree.subsume, base_tree)
        self.build_tree(["subtree2/"])
        sub_tree2 = self.make_branch_and_tree("subtree2")
        self.assertRaises(errors.BadSubsumeSource, sub_tree.subsume, sub_tree2)
        self.build_tree(["tree/subtree/subtree3/"])
        sub_tree3 = self.make_branch_and_tree("tree/subtree/subtree3")
        self.assertRaises(errors.BadSubsumeSource, base_tree.subsume, sub_tree3)



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Directory lookup that uses Launchpad."""

from urllib.parse import urlsplit

from ... import debug, errors, trace, transport
from ...urlutils import InvalidURL, join, split
from .account import get_lp_login
from .uris import LPNET_SERVICE_ROOT

# As breezy.transport.remote may not be loaded yet, make sure bzr+ssh
# is counted as a netloc protocol.
transport.register_urlparse_netloc_protocol("bzr+ssh")
transport.register_urlparse_netloc_protocol("lp")
transport.register_urlparse_netloc_protocol("lp+bzr")


def _requires_launchpad_login(scheme, netloc, path, query, fragment):
    """Does the URL require a Launchpad login in order to be reached?

    The URL is specified by its parsed components, as returned from
    urlsplit.
    """
    return scheme in ("bzr+ssh", "sftp", "git+ssh") and (
        netloc.endswith("launchpad.net") or netloc.endswith("launchpad.test")
    )


def _expand_user(path, url, lp_login):
    if path.startswith("~/"):
        if lp_login is None:
            raise InvalidURL(
                path=url,
                extra='Cannot resolve "~" to your username.'
                ' See "bzr help launchpad-login"',
            )
        path = "~" + lp_login + path[1:]
    return path


def _update_url_scheme(url):
    scheme, netloc, path, query, fragment = urlsplit(url)
    if scheme == "lp+bzr":
        scheme = "lp"
    return url, path


def _resolve_via_api(path, url, api_base_url=LPNET_SERVICE_ROOT):
    from .lp_api import connect_launchpad

    lp = connect_launchpad(api_base_url, version="devel")
    subpaths = []
    lp_branch = None
    git_repo = None
    while path:
        lp_branch = lp.branches.getByPath(path=path)
        git_repo = lp.git_repositories.getByPath(path=path)
        if lp_branch and git_repo:
            target = git_repo.target
            vcs = target.vcs
            trace.warning(
                "Found both a Bazaar branch and a git repository at lp:%s. Using "
                "%s, since that is the projects' default vcs",
                path,
                vcs,
            )
            if vcs == "Git":
                lp_branch = None
            elif vcs == "Bazaar":
                git_repo = None
            else:
                raise errors.BzrError(
                    "Unknown default vcs {} for {}".format(vcs, target)
                )
        if lp_branch or git_repo:
            break
        path, subpath = split(path)
        subpaths.insert(0, subpath)
    if lp_branch:
        return {
            "urls": [
                join(lp_branch.composePublicURL(scheme="bzr+ssh"), *subpaths),
                join(lp_branch.composePublicURL(scheme="http"), *subpaths),
            ]
        }
    elif git_repo:
        return {
            "urls": [
                join(git_repo.git_ssh_url, *subpaths),
                join(git_repo.git_https_url, *subpaths),
            ]
        }
    else:
        raise InvalidURL(f"Unknown Launchpad path: {url}")


def _resolve(url, _lp_login=None):
    """Resolve the base URL for this transport."""
    url, path = _update_url_scheme(url)
    if _lp_login is None:
        _lp_login = get_lp_login()
    path = path.strip("/")
    path = _expand_user(path, url, _lp_login)
    result = _resolve_via_api(path, url)

    if "launchpad" in debug.debug_flags:
        trace.mutter("resolve_lp_path(%r) == %r", url, result)

    _warned_login = False
    for url in result["urls"]:
        scheme, netloc, path, query, fragment = urlsplit(url)
        if _requires_launchpad_login(scheme, netloc, path, query, fragment):
            # Only accept launchpad.net bzr+ssh URLs if we know
            # the user's Launchpad login:
            if _lp_login is not None:
                break
            if _lp_login is None:
                if not _warned_login:
                    trace.warning(
                        "You have not informed bzr of your Launchpad ID, and you must do this to\n"
                        'write to Launchpad or access private data.  See "bzr help launchpad-login".'
                    )
                    _warned_login = True
        else:
            break
    else:
        raise InvalidURL(path=url, extra="no supported schemes")
    return url


class LaunchpadDirectory:
    def look_up(self, name, url, purpose=None):
        """See DirectoryService.look_up."""
        return _resolve(url)


def get_test_permutations():
    # Since this transport doesn't do anything once opened, it's not subjected
    # to the usual transport tests.
    return []



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that all WorkingTree's implement get_file_mtime."""

import os

from breezy import errors, transport
from breezy.tests import TestNotApplicable
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestGetFileMTime(TestCaseWithWorkingTree):
    """Test WorkingTree.get_file_mtime.

    These are more involved because we need to handle files which have been
    renamed, etc.
    """

    def make_basic_tree(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/one"])
        tree.add(["one"])
        return tree

    def test_get_file_mtime_nested(self):
        tree = self.make_basic_tree()
        subtree = self.make_branch_and_tree("tree/sub")
        self.build_tree(["tree/sub/one"])
        subtree.add(["one"])
        subtree.commit("one")
        try:
            tree.add_reference(subtree)
        except errors.UnsupportedOperation:
            raise TestNotApplicable("subtrees not supported")
        tree.commit("sub")

        with tree.lock_read(), subtree.lock_read():
            self.assertEqual(
                tree.get_file_mtime("sub/one"), subtree.get_file_mtime("one")
            )
            self.assertEqual(
                tree.basis_tree().get_file_mtime("sub/one"),
                subtree.basis_tree().get_file_mtime("one"),
            )

    def test_get_file_mtime(self):
        tree = self.make_basic_tree()

        st = os.lstat("tree/one")
        with tree.lock_read():
            mtime_file_id = tree.get_file_mtime("one")
            self.assertIsInstance(mtime_file_id, (float, int))
            self.assertAlmostEqual(st.st_mtime, mtime_file_id)

    def test_after_commit(self):
        """Committing shouldn't change the mtime."""
        tree = self.make_basic_tree()

        st = os.lstat("tree/one")
        tree.commit("one")

        with tree.lock_read():
            mtime = tree.get_file_mtime("one")
            self.assertAlmostEqual(st.st_mtime, mtime)

    def test_get_renamed_time(self):
        """We should handle renamed files."""
        tree = self.make_basic_tree()

        tree.rename_one("one", "two")
        st = os.lstat("tree/two")

        with tree.lock_read():
            mtime = tree.get_file_mtime("two")
            self.assertAlmostEqual(st.st_mtime, mtime)

    def test_get_renamed_in_subdir_time(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/d/", "tree/d/a"])
        tree.add(["d", "d/a"])
        tree.commit("1")

        tree.rename_one("d", "e")

        st = os.lstat("tree/e/a")
        with tree.lock_read():
            mtime = tree.get_file_mtime("e/a")
            self.assertAlmostEqual(st.st_mtime, mtime)

    def test_missing(self):
        tree = self.make_basic_tree()

        os.remove("tree/one")
        with tree.lock_read():
            self.assertRaises(transport.NoSuchFile, tree.get_file_mtime, "one")



# Copyright (C) 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the extra cases that WorkingTree.walkdirs can encounter."""

import os

from breezy.tests.features import SymlinkFeature
from breezy.tests.per_workingtree import TestCaseWithWorkingTree

# tests to write:
# type mismatches - file to link, dir, dir to file, link, link to file, dir


class DirBlock:
    """Object representation of the tuples returned by dirstate."""

    def __init__(
        self,
        tree,
        file_path,
        file_name=None,
        inventory_kind=None,
        stat=None,
        disk_kind="unknown",
    ):
        self.file_path = file_path
        self.abspath = tree.abspath(file_path)
        self.relpath = tree.relpath(file_path)
        if file_name is None:
            file_name = os.path.split(file_path)[-1]
            if len(file_name) == 0:
                file_name = os.path.split(file_path)[-2]
        self.file_name = file_name
        self.inventory_kind = inventory_kind
        self.stat = stat
        self.disk_kind = disk_kind

    def as_tuple(self):
        return (
            self.relpath,
            self.file_name,
            self.disk_kind,
            self.stat,
            self.inventory_kind,
        )

    def as_dir_tuple(self):
        return self.relpath

    def __str__(self):
        return """
file_path      = {!r}
abspath        = {!r}
relpath        = {!r}
file_name      = {!r}
inventory_kind = {!r}
stat           = {!r}
disk_kind      = {!r}""".format(
            self.file_path,
            self.abspath,
            self.relpath,
            self.file_name,
            self.inventory_kind,
            self.stat,
            self.disk_kind,
        )


class TestWalkdirs(TestCaseWithWorkingTree):
    added = "added"
    missing = "missing"
    unknown = "unknown"

    def get_tree(self, file_status, prefix=None):
        tree = self.make_branch_and_tree(".")
        dirblocks = []
        paths = [
            file_status + " file",
            file_status + " dir/",
            file_status + " dir/a file",
            file_status + " empty dir/",
        ]
        self.build_tree(paths)

        def add_dirblock(path, kind):
            dirblock = DirBlock(tree, path)
            if file_status != self.unknown:
                dirblock.inventory_kind = kind
            if file_status != self.missing:
                dirblock.disk_kind = kind
                dirblock.stat = os.lstat(dirblock.relpath)
            dirblocks.append(dirblock)

        add_dirblock(paths[0], "file")
        add_dirblock(paths[1], "directory")
        add_dirblock(paths[2], "file")
        add_dirblock(paths[3], "directory")

        if file_status != self.unknown:
            tree.add(paths)

        if file_status == self.missing:
            # now make the files be missing
            tree.controldir.root_transport.delete(dirblocks[0].relpath)
            tree.controldir.root_transport.delete_tree(dirblocks[1].relpath)
            tree.controldir.root_transport.delete_tree(dirblocks[3].relpath)

        expected_dirblocks = [
            (
                "",
                [dirblocks[1].as_tuple()]
                + (
                    [dirblocks[3].as_tuple()]
                    if (tree.has_versioned_directories() or file_status == self.unknown)
                    else []
                )
                + [dirblocks[0].as_tuple()],
            ),
            (dirblocks[1].as_dir_tuple(), [dirblocks[2].as_tuple()]),
        ]
        if tree.has_versioned_directories() or file_status != self.missing:
            expected_dirblocks.append((dirblocks[3].as_dir_tuple(), []))
        if prefix:
            expected_dirblocks = [
                e
                for e in expected_dirblocks
                if len(e) > 0 and len(e[0]) > 0 and e[0] == prefix
            ]
        return tree, expected_dirblocks

    def _test_walkdir(self, file_status, prefix=""):
        result = []
        tree, expected_dirblocks = self.get_tree(file_status, prefix)
        with tree.lock_read():
            for dirpath, dirblock in tree.walkdirs(prefix):
                result.append((dirpath, list(dirblock)))

        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            result_pos = []
            if len(result) > pos:
                result_pos = result[pos]
            self.assertEqual(item, result_pos)
        self.assertEqual(expected_dirblocks, result)

    def test_walkdir_unknowns(self):
        """Unknown files and directories should be reported by walkdirs."""
        self._test_walkdir(self.unknown)

    def test_walkdir_from_unknown_dir(self):
        """Doing a walkdir when the requested prefix is unknown but on disk."""
        self._test_walkdir(self.unknown, "unknown dir")

    def test_walkdir_missings(self):
        """Missing files and directories should be reported by walkdirs."""
        self._test_walkdir(self.missing)

    def test_walkdir_from_dir(self):
        """Doing a walkdir when the requested prefix is known and on disk."""
        self._test_walkdir(self.added, "added dir")

    def test_walkdir_from_empty_dir(self):
        """Doing a walkdir when the requested prefix is empty dir."""
        self._test_walkdir(self.added, "added empty dir")

    def test_walkdir_from_missing_dir(self):
        """Doing a walkdir when the requested prefix is missing but on disk."""
        self._test_walkdir(self.missing, "missing dir")

    def test_walkdirs_type_changes(self):
        """Walkdir shows the actual kinds on disk and the recorded kinds."""
        self.requireFeature(SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree(".")
        paths = ["file1", "file2", "dir1/", "dir2/"]
        self.build_tree(paths)
        tree.add(paths)
        tt = tree.transform()
        root_transaction_id = tt.trans_id_tree_path("")
        tt.new_symlink("link1", root_transaction_id, "link-target", b"link1")
        tt.new_symlink("link2", root_transaction_id, "link-target", b"link2")
        tt.apply()
        tree.controldir.root_transport.delete_tree("dir1")
        tree.controldir.root_transport.delete_tree("dir2")
        tree.controldir.root_transport.delete("file1")
        tree.controldir.root_transport.delete("file2")
        tree.controldir.root_transport.delete("link1")
        tree.controldir.root_transport.delete("link2")
        changed_paths = ["dir1", "file1/", "link1", "link2/"]
        self.build_tree(changed_paths)
        os.symlink("target", "dir2")
        os.symlink("target", "file2")
        dir1_stat = os.lstat("dir1")
        dir2_stat = os.lstat("dir2")
        file1_stat = os.lstat("file1")
        file2_stat = os.lstat("file2")
        link1_stat = os.lstat("link1")
        link2_stat = os.lstat("link2")
        expected_dirblocks = [
            (
                "",
                [
                    (
                        "dir1",
                        "dir1",
                        "file",
                        dir1_stat,
                        "directory" if tree.has_versioned_directories() else None,
                    ),
                    (
                        "dir2",
                        "dir2",
                        "symlink",
                        dir2_stat,
                        "directory" if tree.has_versioned_directories() else None,
                    ),
                    ("file1", "file1", "directory", file1_stat, "file"),
                    ("file2", "file2", "symlink", file2_stat, "file"),
                    ("link1", "link1", "file", link1_stat, "symlink"),
                    ("link2", "link2", "directory", link2_stat, "symlink"),
                ],
            )
        ]
        if tree.has_versioned_directories():
            expected_dirblocks.extend([("dir1", []), ("dir2", [])])
        expected_dirblocks.extend(
            [
                ("file1", []),
                ("link2", []),
            ]
        )
        with tree.lock_read():
            result = list(tree.walkdirs())
        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            self.assertEqual(item, result[pos])
        self.assertEqual(len(expected_dirblocks), len(result))

    def test_walkdirs_type_changes_wo_symlinks(self):
        # similar to test_walkdirs_type_changes
        # but don't use symlinks for safe testing on win32
        tree = self.make_branch_and_tree(".")
        paths = ["file1", "dir1/"]
        self.build_tree(paths)
        tree.add(paths)
        tree.controldir.root_transport.delete_tree("dir1")
        tree.controldir.root_transport.delete("file1")
        changed_paths = ["dir1", "file1/"]
        self.build_tree(changed_paths)
        dir1_stat = os.lstat("dir1")
        file1_stat = os.lstat("file1")
        if tree.has_versioned_directories():
            expected_dirblocks = [
                (
                    "",
                    [
                        ("dir1", "dir1", "file", dir1_stat, "directory"),
                        ("file1", "file1", "directory", file1_stat, "file"),
                    ],
                ),
                ("dir1", []),
                ("file1", []),
            ]
        else:
            expected_dirblocks = [
                (
                    "",
                    [
                        ("dir1", "dir1", "file", dir1_stat, None),
                        ("file1", "file1", "directory", file1_stat, "file"),
                    ],
                ),
                ("file1", []),
            ]
        with tree.lock_read():
            result = list(tree.walkdirs())
        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            self.assertEqual(item, result[pos])
        self.assertEqual(len(expected_dirblocks), len(result))



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for test feature dependencies."""

import sys

from .. import plugin as _mod_plugin
from .. import symbol_versioning, tests
from . import features


class TestFeature(tests.TestCase):
    def test_caching(self):
        """Feature._probe is called by the feature at most once."""

        class InstrumentedFeature(features.Feature):
            def __init__(self):
                super().__init__()
                self.calls = []

            def _probe(self):
                self.calls.append("_probe")
                return False

        feature = InstrumentedFeature()
        feature.available()
        self.assertEqual(["_probe"], feature.calls)
        feature.available()
        self.assertEqual(["_probe"], feature.calls)

    def test_named_str(self):
        """Feature.__str__ should thunk to feature_name()."""

        class NamedFeature(features.Feature):
            def feature_name(self):
                return "symlinks"

        feature = NamedFeature()
        self.assertEqual("symlinks", str(feature))

    def test_default_str(self):
        """Feature.__str__ should default to __class__.__name__."""

        class NamedFeature(features.Feature):
            pass

        feature = NamedFeature()
        self.assertEqual("NamedFeature", str(feature))


class TestUnavailableFeature(tests.TestCase):
    def test_access_feature(self):
        feature = features.Feature()
        exception = tests.UnavailableFeature(feature)
        self.assertIs(feature, exception.args[0])


# Although this was deprecated a long time ago, please keep it here because
# it's really just a test fixture for test-feature deprecation.
simple_thunk_feature = features._CompatabilityThunkFeature(
    symbol_versioning.deprecated_in((2, 1, 0)),
    "breezy.tests.test_features",
    "simple_thunk_feature",
    "UnicodeFilenameFeature",
    replacement_module="breezy.tests.features",
)


class Test_CompatibilityFeature(tests.TestCase):
    def test_does_thunk(self):
        res = self.callDeprecated(
            [
                "breezy.tests.test_features.simple_thunk_feature "
                "was deprecated in version 2.1.0. "
                "Use breezy.tests.features.UnicodeFilenameFeature instead."
            ],
            simple_thunk_feature.available,
        )
        self.assertEqual(features.UnicodeFilenameFeature.available(), res)

    def test_reports_correct_location(self):
        a_feature = features._CompatabilityThunkFeature(
            symbol_versioning.deprecated_in((2, 1, 0)),
            "breezy.tests.test_features",
            "a_feature",
            "UnicodeFilenameFeature",
            replacement_module="breezy.tests.features",
        )

        def test_caller(message, category=None, stacklevel=1):
            # Find ourselves back from the right frame
            caller = sys._getframe(stacklevel)
            reported_file = caller.f_globals["__file__"]
            reported_lineno = caller.f_lineno
            self.assertEqual(__file__, reported_file)
            # The call we're tracking occurred the line after we grabbed the
            # lineno.
            self.assertEqual(self.lineno + 1, reported_lineno)

        self.overrideAttr(symbol_versioning, "warn", test_caller)
        # Grab the current lineno
        self.lineno = sys._getframe().f_lineno
        self.requireFeature(a_feature)


class TestModuleAvailableFeature(tests.TestCase):
    def test_available_module(self):
        feature = features.ModuleAvailableFeature("breezy.tests")
        self.assertEqual("breezy.tests", feature.module_name)
        self.assertEqual("breezy.tests", str(feature))
        self.assertTrue(feature.available())
        self.assertIs(tests, feature.module)

    def test_unavailable_module(self):
        feature = features.ModuleAvailableFeature("breezy.no_such_module_exists")
        self.assertEqual("breezy.no_such_module_exists", str(feature))
        self.assertFalse(feature.available())
        self.assertIs(None, feature.module)


class TestPluginLoadedFeature(tests.TestCase):
    def test_available_plugin(self):
        plugins = _mod_plugin.plugins()
        if not plugins:
            self.skipTest("no plugins available to test with")
        a_plugin_name = next(iter(plugins))
        feature = features.PluginLoadedFeature(a_plugin_name)
        self.assertEqual(a_plugin_name, feature.plugin_name)
        self.assertEqual(a_plugin_name + " plugin", str(feature))
        self.assertTrue(feature.available())

    def test_unavailable_plugin(self):
        feature = features.PluginLoadedFeature("idontexist")
        self.assertEqual("idontexist plugin", str(feature))
        self.assertFalse(feature.available())
        self.assertIs(None, feature.plugin)


class TestUnicodeFilenameFeature(tests.TestCase):
    def test_probe_passes(self):
        """UnicodeFilenameFeature._probe passes."""
        # We can't test much more than that because the behaviour depends
        # on the platform.
        features.UnicodeFilenameFeature._probe()


class TestBackslashFilenameFeature(tests.TestCase):
    def test_probe_passes(self):
        """BackslashFilenameFeature._probe passes."""
        # We can't test much more than that because the behaviour depends
        # on the platform.
        features.BackslashFilenameFeature._probe()



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for handling of ignore files."""

import os
from io import BytesIO

from .. import bedding, ignores
from . import TestCase, TestCaseInTempDir, TestCaseWithTransport


class TestParseIgnoreFile(TestCase):
    def test_parse_fancy(self):
        ignored = ignores.parse_ignore_file(
            BytesIO(
                b"./rootdir\n"
                b"randomfile*\n"
                b"path/from/ro?t\n"
                b"unicode\xc2\xb5\n"  # u'\xb5'.encode('utf8')
                b"dos\r\n"
                b"\n"  # empty line
                b"#comment\n"
                b" xx \n"  # whitespace
                b"!RE:^\\.z.*\n"
                b"!!./.zcompdump\n"
            )
        )
        self.assertEqual(
            {
                "./rootdir",
                "randomfile*",
                "path/from/ro?t",
                "unicode\xb5",
                "dos",
                " xx ",
                "!RE:^\\.z.*",
                "!!./.zcompdump",
            },
            ignored,
        )

    def test_parse_empty(self):
        ignored = ignores.parse_ignore_file(BytesIO(b""))
        self.assertEqual(set(), ignored)

    def test_parse_non_utf8(self):
        """Lines with non utf 8 characters should be discarded."""
        ignored = ignores.parse_ignore_file(
            BytesIO(b"utf8filename_a\ninvalid utf8\x80\nutf8filename_b\n")
        )
        self.assertEqual(
            {
                "utf8filename_a",
                "utf8filename_b",
            },
            ignored,
        )


class TestUserIgnores(TestCaseInTempDir):
    def test_create_if_missing(self):
        # $HOME should be set to '.'
        ignore_path = bedding.user_ignore_config_path()
        self.assertPathDoesNotExist(ignore_path)
        user_ignores = ignores.get_user_ignores()
        self.assertEqual(set(ignores.USER_DEFAULTS), user_ignores)

        self.assertPathExists(ignore_path)
        with open(ignore_path, "rb") as f:
            entries = ignores.parse_ignore_file(f)
        self.assertEqual(set(ignores.USER_DEFAULTS), entries)

    def test_create_with_intermediate_missing(self):
        # $HOME should be set to '.'
        ignore_path = bedding.user_ignore_config_path()
        self.assertPathDoesNotExist(ignore_path)
        os.mkdir("empty-home")

        config_path = os.path.join(self.test_dir, "empty-home", "foo", ".config")
        self.overrideEnv("BRZ_HOME", config_path)
        self.assertPathDoesNotExist(config_path)

        user_ignores = ignores.get_user_ignores()
        self.assertEqual(set(ignores.USER_DEFAULTS), user_ignores)

        ignore_path = bedding.user_ignore_config_path()
        self.assertPathDoesNotExist(ignore_path)

    def test_use_existing(self):
        patterns = ["*.o", "*.py[co]", "\xe5*"]
        ignores._set_user_ignores(patterns)

        user_ignores = ignores.get_user_ignores()
        self.assertEqual(set(patterns), user_ignores)

    def test_use_empty(self):
        ignores._set_user_ignores([])
        ignore_path = bedding.user_ignore_config_path()
        self.check_file_contents(ignore_path, b"")

        self.assertEqual(set(), ignores.get_user_ignores())

    def test_set(self):
        patterns = ["*.py[co]", "*.py[oc]"]
        ignores._set_user_ignores(patterns)

        self.assertEqual(set(patterns), ignores.get_user_ignores())

        patterns = ["vim", "*.swp"]
        ignores._set_user_ignores(patterns)
        self.assertEqual(set(patterns), ignores.get_user_ignores())

    def test_add(self):
        """Test that adding will not duplicate ignores."""
        # Create an empty file
        ignores._set_user_ignores([])

        patterns = ["foo", "./bar", "b\xe5z"]
        added = ignores.add_unique_user_ignores(patterns)
        self.assertEqual(patterns, added)
        self.assertEqual(set(patterns), ignores.get_user_ignores())

    def test_add_directory(self):
        """Test that adding a directory will strip any trailing slash."""
        # Create an empty file
        ignores._set_user_ignores([])

        in_patterns = ["foo/", "bar/", "baz\\"]
        added = ignores.add_unique_user_ignores(in_patterns)
        out_patterns = [x.rstrip("/\\") for x in in_patterns]
        self.assertEqual(out_patterns, added)
        self.assertEqual(set(out_patterns), ignores.get_user_ignores())

    def test_add_unique(self):
        """Test that adding will not duplicate ignores."""
        ignores._set_user_ignores(["foo", "./bar", "b\xe5z", "dir1/", "dir3\\"])

        added = ignores.add_unique_user_ignores(
            ["xxx", "./bar", "xxx", "dir1/", "dir2/", "dir3\\"]
        )
        self.assertEqual(["xxx", "dir2"], added)
        self.assertEqual(
            {"foo", "./bar", "b\xe5z", "xxx", "dir1", "dir2", "dir3"},
            ignores.get_user_ignores(),
        )


class TestRuntimeIgnores(TestCase):
    def setUp(self):
        super().setUp()

        # For the purposes of these tests, we must have no
        # runtime ignores
        self.overrideAttr(ignores, "_runtime_ignores", set())

    def test_add(self):
        """Test that we can add an entry to the list."""
        self.assertEqual(set(), ignores.get_runtime_ignores())

        ignores.add_runtime_ignores(["foo"])
        self.assertEqual({"foo"}, ignores.get_runtime_ignores())

    def test_add_duplicate(self):
        """Adding the same ignore twice shouldn't add a new entry."""
        ignores.add_runtime_ignores(["foo", "bar"])
        self.assertEqual({"foo", "bar"}, ignores.get_runtime_ignores())

        ignores.add_runtime_ignores(["bar"])
        self.assertEqual({"foo", "bar"}, ignores.get_runtime_ignores())


class TestTreeIgnores(TestCaseWithTransport):
    def assertPatternsEquals(self, patterns):
        with open(".bzrignore", "rb") as f:
            contents = f.read().decode("utf-8").splitlines()
        self.assertEqual(sorted(patterns), sorted(contents))

    def test_new_file(self):
        tree = self.make_branch_and_tree(".")
        ignores.tree_ignores_add_patterns(tree, ["myentry"])
        self.assertTrue(tree.has_filename(".bzrignore"))
        self.assertPatternsEquals(["myentry"])

    def test_add_to_existing(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([(".bzrignore", b"myentry1\n")])
        tree.add([".bzrignore"])
        ignores.tree_ignores_add_patterns(tree, ["myentry2", "foo"])
        self.assertPatternsEquals(["myentry1", "myentry2", "foo"])

    def test_adds_ending_newline(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([(".bzrignore", b"myentry1")])
        tree.add([".bzrignore"])
        ignores.tree_ignores_add_patterns(tree, ["myentry2"])
        self.assertPatternsEquals(["myentry1", "myentry2"])
        with open(".bzrignore") as f:
            text = f.read()
        self.assertTrue(text.endswith(("\r\n", "\n", "\r")))

    def test_does_not_add_dupe(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([(".bzrignore", b"myentry\n")])
        tree.add([".bzrignore"])
        ignores.tree_ignores_add_patterns(tree, ["myentry"])
        self.assertPatternsEquals(["myentry"])

    def test_non_ascii(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([(".bzrignore", "myentry\u1234\n".encode())])
        tree.add([".bzrignore"])
        ignores.tree_ignores_add_patterns(tree, ["myentry\u5678"])
        self.assertPatternsEquals(["myentry\u1234", "myentry\u5678"])

    def test_crlf(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([(".bzrignore", b"myentry1\r\n")])
        tree.add([".bzrignore"])
        ignores.tree_ignores_add_patterns(tree, ["myentry2", "foo"])
        with open(".bzrignore", "rb") as f:
            self.assertEqual(f.read(), b"myentry1\r\nmyentry2\r\nfoo\r\n")
        self.assertPatternsEquals(["myentry1", "myentry2", "foo"])



# Copyright (C) 2007 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.bzr import inventorytree
from breezy.tests import TestNotApplicable
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestNestedSupport(TestCaseWithWorkingTree):
    def make_branch_and_tree(self, path):
        tree = TestCaseWithWorkingTree.make_branch_and_tree(self, path)
        if not tree.supports_tree_reference():
            raise TestNotApplicable("Tree references not supported")
        return tree

    def test_set_get_inventory_tree_reference(self):
        """This tests that setting a tree reference is persistent."""
        tree = self.make_branch_and_tree(".")
        if not isinstance(tree, inventorytree.InventoryTree):
            raise TestNotApplicable("not an inventory tree")
        transform = tree.transform()
        trans_id = transform.new_directory("reference", transform.root, b"subtree-id")
        transform.set_tree_reference(b"subtree-revision", trans_id)
        transform.apply()
        tree = tree.controldir.open_workingtree()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            b"subtree-revision",
            tree.root_inventory.get_entry(b"subtree-id").reference_revision,
        )

    def test_extract_while_locked(self):
        tree = self.make_branch_and_tree(".")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree(["subtree/"])
        tree.add(["subtree"])
        tree.extract("subtree")

    def prepare_with_subtree(self):
        tree = self.make_branch_and_tree(".")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        subtree = self.make_branch_and_tree("subtree")
        subtree.commit("dummy")
        tree.add(["subtree"])
        return tree

    def test_comparison_data_does_not_autodetect_subtree(self):
        tree = self.prepare_with_subtree()
        (path, versioned, kind, ie) = list(tree.list_files("subtree"))[0]
        self.assertEqual("directory", tree._comparison_data(ie, "subtree")[0])

    def test_may_not_autodetect_subtree(self):
        tree = self.prepare_with_subtree()
        self.assertIn(tree.kind("subtree"), ("directory", "tree-reference"))

    def test_iter_entries_by_dir_autodetects_subtree(self):
        tree = self.prepare_with_subtree()
        path, ie = next(tree.iter_entries_by_dir(specific_files=["subtree"]))
        self.assertEqual("tree-reference", ie.kind)



# Copyright (C) 2005-2013, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for bzr.

These check that it behaves properly when it's invoked through the regular
command-line interface. This doesn't actually run a new interpreter but
rather starts again from the run_brz function.
"""

from breezy import tests


def load_tests(loader, basic_tests, pattern):
    suite = loader.suiteClass()
    # add the tests for this module
    suite.addTests(basic_tests)

    prefix = "breezy.tests.blackbox."
    testmod_names = [
        "test_add",
        "test_added",
        "test_alias",
        "test_aliases",
        "test_ancestry",
        "test_annotate",
        "test_bisect",
        "test_big_file",
        "test_branch",
        "test_branches",
        "test_break_lock",
        "test_bound_branches",
        "test_bundle_info",
        "test_cat",
        "test_cat_revision",
        "test_check",
        "test_checkout",
        "test_clean_tree",
        "test_clone",
        "test_command_encoding",
        "test_commit",
        "test_config",
        "test_conflicts",
        "test_cp",
        "test_debug",
        "test_deleted",
        "test_diff",
        "test_exceptions",
        "test_export",
        "test_export_pot",
        "test_fetch_ghosts",
        "test_filesystem_cicp",
        "test_filtered_view_ops",
        "test_find_merge_base",
        "test_help",
        "test_hooks",
        "test_import",
        "test_ignore",
        "test_ignored",
        "test_info",
        "test_init",
        "test_inventory",
        "test_join",
        "test_link_tree",
        "test_locale",
        "test_log",
        "test_logformats",
        "test_lookup_revision",
        "test_ls",
        "test_lsprof",
        "test_merge",
        "test_merge_directive",
        "test_missing",
        "test_mkdir",
        "test_modified",
        "test_mv",
        "test_nick",
        "test_non_ascii",
        "test_outside_wt",
        "test_pack",
        "test_patch",
        "test_ping",
        "test_plugins",
        "test_pull",
        "test_push",
        "test_reconcile",
        "test_reconfigure",
        "test_reference",
        "test_remerge",
        "test_remove",
        "test_re_sign",
        "test_remember_option",
        "test_remove_tree",
        "test_repair_workingtree",
        "test_resolve",
        "test_resolve_location",
        "test_revert",
        "test_revno",
        "test_revision_history",
        "test_revision_info",
        "test_rmbranch",
        "test_script",
        "test_selftest",
        "test_send",
        "test_serve",
        "test_shared_repository",
        "test_shell_complete",
        "test_shelve",
        "test_sign_my_commits",
        "test_verify_signatures",
        "test_split",
        "test_status",
        "test_switch",
        "test_tags",
        "test_testament",
        "test_too_much",
        "test_uncommit",
        "test_unknowns",
        "test_update",
        "test_upgrade",
        "test_version",
        "test_version_info",
        "test_versioning",
        "test_view",
        "test_whoami",
    ]
    # add the tests for the sub modules
    suite.addTests(
        loader.loadTestsFromModuleNames(
            [prefix + module_name for module_name in testmod_names]
        )
    )
    return suite


class ExternalBase(tests.TestCaseWithTransport):
    """Don't use this class anymore, use TestCaseWithTransport or similar."""
